@@ -8,7 +8,7 @@ from torch.optim import Optimizer
 import matplotlib.pyplot as plt
 import itertools
 import time
-from typing import TypedDict, Union, List
+from typing import TypedDict, Union, List, Tuple
 
 # documentation
 ModuleOutput = TypedDict('ModuleOutput', {'loss': torch.Tensor, 'scores': torch.Tensor}, total=False)
@@ -88,8 +88,8 @@ def make_reproducible(seed: int = 1):
     torch.cuda.manual_seed(seed)
     np.random.seed(seed)
     random.seed(seed)
-    torch.cudnn.deterministic = True
-    torch.cudnn.benchmark = False
+    torch.backends.cudnn.deterministic = True
+    torch.backends.cudnn.benchmark = False
 
     torch.set_printoptions(sci_mode=False)
     torch.set_printoptions(threshold=100_000)
@@ -102,40 +102,58 @@ class Setup:
 
     :param loader_train: DataLoader for the training set
     :type loader_train: torch.utils.data.DataLoader
+
     :param loader_val: DataLoader for the validation set
     :type loader_val: torch.utils.data.DataLoader
+
     :param loader_test: DataLoader for the test set
     :type loader_test: torch.utils.data.DataLoader
+
     :param device: device to run the training process on.
     :type device: str
+
     :param monitor_n_losses: the number of steps after which the loss slope is printed.
     :type monitor_n_losses: int
+
     :param checkpoint_initial: path to the initial checkpoint.
     :type checkpoint_initial: str
+
     :param checkpoint_running: path to the running checkpoint.
     :type checkpoint_running: str
+
     :param checkpoint_final: path to the final checkpoint.
     :type checkpoint_final: str
+
     :param lrrt_n_batches: number of batches used in the learning rate range test.
     :type lrrt_n_batches: int
+
     :param lrrt_slope_desired: exclusive border used in the learning rate range test.
     :type lrrt_slope_desired: float
+
     :param lrrt_max_decays: maximum number of learning rate decays performed in the learning rate range test.
     :type lrrt_max_decays: int
+
     :param lrrt_decay: decay factor used in the learning rate range test.
     :type lrrt_decay: float
+
     :param lrrt_initial_candidates: initial candidate learning rates used in the learning rate range test.
     :type lrrt_initial_candidates: np.ndarray
+
     :param overkill_initial_lr: initial learning rate used in overkill training
     :type overkill_initial_lr: float
+
     :param overkill_decay: learning rate decay factor used in overkill training
     :type overkill_decay: float
+
     :param overkill_max_violations: maximum number of violations allowed in overkill training
     :type overkill_max_violations: int
+
     :param overkill_max_decays: maximum number of learning rate decays allowed in overkill training
     :type overkill_max_decays: int
+
     :param es_max_violations: maximum number of violations allowed in early stopping
     :type es_max_violations: int
+
     :param optimizer_class: optimizer class used for the training process
     :type optimizer_class: type[torch.optim.Optimizer]
     """
@@ -201,8 +219,7 @@ class BatchHandler:
 
     def forward_batch(self,
                       module: Module,
-                      batch: List[Union[torch.Tensor, List[torch.Tensor]],
-                                  Union[torch.Tensor, List[torch.Tensor]]]) -> ModuleOutput:
+                      batch: Tuple[Union[torch.Tensor, List[torch.Tensor]], Union[torch.Tensor, List[torch.Tensor]]]) -> ModuleOutput:
         """
         performs a forward pass with a given module and batch.
 
@@ -210,7 +227,7 @@ class BatchHandler:
         :type module: Module
 
         :param batch: batch containing x and y
-        :type batch: List[Union[torch.Tensor, List[torch.Tensor]], Union[torch.Tensor, List[torch.Tensor]]]
+        :type batch: Tuple[Union[torch.Tensor, List[torch.Tensor]], Union[torch.Tensor, List[torch.Tensor]]]
 
         :return: output of the forward pass
         :rtype: ModuleOutput
@@ -224,8 +241,7 @@ class BatchHandler:
     def train_batch(self,
                     module: Module,
                     optimizer: Optimizer,
-                    batch: List[Union[torch.Tensor, List[torch.Tensor]],
-                                Union[torch.Tensor, List[torch.Tensor]]],
+                    batch: Tuple[Union[torch.Tensor, List[torch.Tensor]], Union[torch.Tensor, List[torch.Tensor]]],
                     freeze_pretrained: bool = False) -> float:
         """
         train on one batch
@@ -237,7 +253,7 @@ class BatchHandler:
         :type optimizer: Optimizer
 
         :param batch: batch containing x, y
-        :type batch: List[Union[torch.Tensor, List[torch.Tensor]], Union[torch.Tensor, List[torch.Tensor]]]
+        :type batch: Tuple[Union[torch.Tensor, List[torch.Tensor]], Union[torch.Tensor, List[torch.Tensor]]]
 
         :param freeze_pretrained: determines if pretrained layers are frozen
         :type freeze_pretrained: bool
@@ -259,8 +275,7 @@ class BatchHandler:
 
     def loss_batch_eval(self,
                         module: Module,
-                        batch: List[Union[torch.Tensor, List[torch.Tensor]],
-                                    Union[torch.Tensor, List[torch.Tensor]]]) -> float:
+                        batch: Tuple[Union[torch.Tensor, List[torch.Tensor]], Union[torch.Tensor, List[torch.Tensor]]]) -> float:
         """
         calculates the loss on one batch in evaluation mode
 
@@ -268,7 +283,7 @@ class BatchHandler:
         :type module: Module
 
         :param batch: batch containing x, y
-        :type batch: List[Union[torch.Tensor, List[torch.Tensor]], Union[torch.Tensor, List[torch.Tensor]]]
+        :type batch: Tuple[Union[torch.Tensor, List[torch.Tensor]], Union[torch.Tensor, List[torch.Tensor]]]
 
         :return: float representation of the batch loss
         :rtype: float
@@ -282,7 +297,7 @@ class BatchHandler:
                         optimizer: Optimizer,
                         n_batches: int,
                         loader: DataLoader,
-                        freeze_pretrained: bool) -> List[List, float, float]:
+                        freeze_pretrained: bool) -> Tuple[List[float], float, float]:
         """
         trains a given module on n batches of a given data loader.
 
@@ -317,11 +332,11 @@ class BatchHandler:
                 slope_last, _ = np.polyfit(x=np.arange(len(losses_last)), y=losses_last, deg=1)
                 print("iter", train_iter + 1, "mean loss", losses_last.mean(), "loss slope", slope_last)
         slope_total, bias_total = np.polyfit(x=np.arange(len(losses)), y=losses, deg=1)
-        return [losses, float(slope_total), float(bias_total)]
+        return losses, float(slope_total), float(bias_total)
 
     def lrrt(self,
              loader: DataLoader,
-             freeze_pretrained: bool = False) -> List[float, float]:
+             freeze_pretrained: bool = False) -> Tuple[float, float]:
         """
         Learning Rate Range Test; basic idea:
         for each learning rate in a set of learning rate candidates:
@@ -371,7 +386,40 @@ class BatchHandler:
         if not slope_desired_found:
             print("lr with desired loss slope", self.setup.lrrt_slope_desired, "not found. using approx best lr")
         print("best loss slope", slope_best_total, "best lr", lr_best_total)
-        return [lr_best_total, slope_best_total]
+        return lr_best_total, slope_best_total
+
+    def loss_epoch_eval(self, module: Module, loader_eval: DataLoader) -> float:
+        """
+        calculates the epoch loss in evaluation mode
+
+        :param module: module to be evaluated
+        :type module: Module
+
+        :param loader_eval: data loader used for evaluation
+        :type loader_eval: DataLoader
+
+        :return: epoch loss
+        :rtype: float
+        """
+        batch_losses = np.zeros(len(loader_eval))
+        for batch_nr, batch in enumerate(loader_eval):
+            batch_losses[batch_nr] = self.loss_batch_eval(module=module, batch=batch)
+        return float(batch_losses.mean())
+
+    def losses_epoch_eval(self, module: Module) -> Tuple[float, float]:
+        """
+        wrapper for loss_epoch_eval
+        calculates epoch loss for training data and validation data
+
+        :param module: module to be evaluated
+        :type module: Module
+
+        :return: training epoch loss, validation epoch loss
+        :rtype: List[float, float]
+        """
+        loss_epoch_train = self.loss_epoch_eval(module=module, loader_eval=self.setup.loader_train)
+        loss_epoch_val = self.loss_epoch_eval(module=module, loader_eval=self.setup.loader_val)
+        return loss_epoch_train, loss_epoch_val
 
 
 class Debugger(BatchHandler):
@@ -383,12 +431,11 @@ class Debugger(BatchHandler):
 
     def overfit_batch(self,
                       module: Module,
-                      batch_debug: List[Union[torch.Tensor, List[torch.Tensor]],
-                                        Union[torch.Tensor, List[torch.Tensor]]],
+                      batch_debug: Tuple[Union[torch.Tensor, List[torch.Tensor]], Union[torch.Tensor, List[torch.Tensor]]],
                       n_iters: int,
                       lrrt_loader: DataLoader = None,
                       lr: float = 1e-5,
-                      freeze_pretrained: bool = False) -> List[Module, List[float]]:
+                      freeze_pretrained: bool = False) -> Tuple[Module, List[float]]:
         """
         overfits one batch to determine if the module can learn.
         used to determine significant bugs in the module structure
@@ -398,7 +445,7 @@ class Debugger(BatchHandler):
         :type module: Module
 
         :param batch_debug: single batch to debug on
-        :type batch_debug: List[Union[torch.Tensor, List[torch.Tensor]], Union[torch.Tensor, List[torch.Tensor]]]
+        :type batch_debug: Tuple[Union[torch.Tensor, List[torch.Tensor]], Union[torch.Tensor, List[torch.Tensor]]]
 
         :param n_iters: determines how many iterations the module is trained on the debug batch
         :type n_iters: int
@@ -413,7 +460,7 @@ class Debugger(BatchHandler):
         :type freeze_pretrained: bool
 
         :return: module, batch losses
-        :rtype: List[Module, List[float]]
+        :rtype: Tuple[Module, List[float]]
         """
         if lrrt_loader is not None:
             lr, _ = self.lrrt(loader=lrrt_loader, freeze_pretrained=freeze_pretrained)
@@ -424,7 +471,7 @@ class Debugger(BatchHandler):
         for _ in range(n_iters):
             losses.append(self.train_batch(module=module, optimizer=optimizer, batch=batch_debug,
                                            freeze_pretrained=freeze_pretrained))
-        return [module, losses]
+        return module, losses
 
 
 class Visualizer:
@@ -450,7 +497,7 @@ class Visualizer:
     @staticmethod
     def lines_multiplot(lines: List[List[float]],
                         title: str,
-                        multiplot_labels: List[List[str]],
+                        multiplot_labels: List[str],
                         y_label: str,
                         x_label: str,
                         file_name: str):
@@ -488,7 +535,7 @@ class Visualizer:
     def lines_subplot(self,
                       lines: List[List[float]],
                       title: str,
-                      subplot_titles: List[List[str]],
+                      subplot_titles: List[str],
                       y_label: str,
                       x_label: str,
                       file_name: str):
@@ -535,42 +582,9 @@ class Trainer(BatchHandler):
     def __init__(self, setup: Setup):
         super(Trainer, self).__init__(setup=setup)
 
-    def loss_epoch_eval(self, module: Module, loader_eval: DataLoader) -> float:
-        """
-        calculates the epoch loss in evaluation mode
-
-        :param module: module to be evaluated
-        :type module: Module
-
-        :param loader_eval: data loader used for evaluation
-        :type loader_eval: DataLoader
-
-        :return: epoch loss
-        :rtype: float
-        """
-        batch_losses = np.zeros(len(loader_eval))
-        for batch_nr, batch in enumerate(loader_eval):
-            batch_losses[batch_nr] = self.loss_batch_eval(module=module, batch=batch)
-        return float(batch_losses.mean())
-
-    def losses_epoch_eval(self, module: Module) -> List[float, float]:
-        """
-        wrapper for loss_epoch_eval
-        calculates epoch loss for training data and validation data
-
-        :param module: module to be evaluated
-        :type module: Module
-
-        :return: training epoch loss, validation epoch loss
-        :rtype: List[float, float]
-        """
-        loss_epoch_train = self.loss_epoch_eval(module=module, loader_eval=self.setup.loader_train)
-        loss_epoch_val = self.loss_epoch_eval(module=module, loader_eval=self.setup.loader_val)
-        return [loss_epoch_train, loss_epoch_val]
-
     def train(self,
               max_epochs: int,
-              freeze_pretrained: bool = False) -> List[Module, List[float], List[float], List[float]]:
+              freeze_pretrained: bool = False) -> Tuple[Module, List[float], List[float], List[float]]:
         """
         training procedure
         can perform learning rate range test and early stopping if specified in the setup.
@@ -633,9 +647,9 @@ class Trainer(BatchHandler):
                 if es_violations == self.setup.es_max_violations:
                     print("early stopping")
                     break
-        return [load_secure(self.setup.checkpoint_final), best_lrs, losses_train, losses_val]
+        return load_secure(self.setup.checkpoint_final), best_lrs, losses_train, losses_val
 
-    def train_overkill_epoch(self, lr_initial: float, freeze_pretrained: bool) -> List[List[float], float, int]:
+    def train_overkill_epoch(self, lr_initial: float, freeze_pretrained: bool) -> Tuple[List[float], float, int]:
         """
         one epoch of overkill training
         for each batch, overkill training decays the learning rate until an improvement
@@ -672,11 +686,11 @@ class Trainer(BatchHandler):
                     violation_counter -= 1
                     break
                 lr = lr * self.setup.overkill_decay
-        return [best_lrs, loss_val_last, violation_counter]
+        return best_lrs, loss_val_last, violation_counter
 
     def train_overkill(self,
                        max_epochs: int,
-                       freeze_pretrained: bool = False) -> List[Module, List[float], List[float], List[float]]:
+                       freeze_pretrained: bool = False) -> Tuple[Module, List[float], List[float], List[float]]:
         """
         performs multiple epochs of overkill training.
         early stops when less than a desired amount of updates was performed per epoch.
@@ -715,4 +729,4 @@ class Trainer(BatchHandler):
                 print("too few updates in this epoch, convergence")
                 break
         torch.save(module, self.setup.checkpoint_final)
-        return [module, list(itertools.chain(*best_lrs)), losses_train, losses_val]
+        return module, list(itertools.chain(*best_lrs)), losses_train, losses_val
